# allow users to do
#     from autoclass import xxx
from autoclass.autoargs import *
from autoclass.autoclass import *
from autoclass.autoprops import *
from autoclass.autodict import *
from autoclass.var_checker import *

# allow users to do
#     import autoclass as ac
__all__ = ['autoclass', 'autoargs', 'autoprops', 'autodict', 'var_checker']
